from .jragbeer_common_db import backup_databases as backup_databases
from .jragbeer_common_db import db_drop_duplicates as db_drop_duplicates
from .jragbeer_common_db import \
    sql_table_drop_duplicates_single as sql_table_drop_duplicates_single
from .jragbeer_common_db import sql_table_drop_duplicates as sql_table_drop_duplicates
from .jragbeer_common_db import pg_backup_database as pg_backup_database
from .jragbeer_common_db import pg_restore_database as pg_restore_database



import datetime
import os
import subprocess
from typing import Optional

import numpy as np
import pandas as pd
import sqlalchemy
from tqdm import tqdm

from ..azure.jragbeer_common_azure import adls_upload_file
from ..common.jragbeer_common_data_eng import (dagster_logger, data_path,
                                               error_handling, today)
from ..dask.jragbeer_common_dask import process_list_with_dask
from ..ubuntu.jragbeer_common_ubuntu import execute_cmd_ubuntu_sudo, execute_cmd_ubuntu_normal

# MYSQL
def mysql_create_database(host, port, username, password, database_name):
    """
    Create a MySQL database using the mysqladmin command.

    Args:
        host (str): The MySQL server hostname or IP address.
        port (int): The MySQL server port.
        username (str): The MySQL username for authentication.
        password (str): The MySQL password for authentication.
        database_name (str): The name of the database to be created.

    Returns:
        bool: True if the database is created successfully, False otherwise.
    """
    try:
        # Construct the mysqladmin command to create the database
        cmd = [
            "/bin/mysqladmin",
            f"--host={host}",
            f"--port={port}",
            f"--user={username}",
            f"--password={password}",
            "create",
            database_name,
        ]

        # Use subprocess.Popen to execute the command
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        if process.returncode == 0:
            print(f"Database '{database_name}' created successfully.")
            return True
        else:
            print(f"Error creating database '{database_name}': {stderr.decode()}")
            return False
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        return False
def mysql_backup_database(database_name: str):
    dagster_logger.info(f"{database_name} dump starting now")
    t = [r"mysqldump", rf"-h{os.getenv('local_db_address')}", rf"-u{os.getenv('local_db_username')}",
         rf"-p{os.getenv('local_db_password')}",
         rf"{database_name}", rf"--result-file={data_path + database_name}_dump_latest.sql"]
    print(t)
    execute_cmd_ubuntu_sudo(t)
    dagster_logger.info(f"{database_name} complete in {datetime.datetime.now() - today} ")
def mysql_import_dump(host, port, username, password, database_name, dump_file_path):
    """
    Imports a MySQL database dump into a specified database.

    Args:
        host (str): The MySQL server hostname or IP address.
        port (int): The MySQL server port.
        username (str): The MySQL username for authentication.
        password (str): The MySQL password for authentication.
        database_name (str): The name of the database where the dump will be imported.
        dump_file_path (str): The path to the MySQL dump file to import.

    Returns:
        bool: True if the import was successful, False otherwise.
    """
    try:
        # Construct the MySQL command to import the dump
        cmd = [
            "/bin/mysql",
            f"--host={host}",
            f"--port={port}",
            f"--user={username}",
            f"--password={password}",
            f"--database={database_name}",
            "--execute=source " + dump_file_path,
        ]

        # Execute the command
        subprocess.run(cmd, check=True, shell=True)
        return True
    except subprocess.CalledProcessError as e:
        print(f"Error importing MySQL dump: {e}")
        return False
def mysql_restore_databases():
    for finance_db in [
        "finance_1mins",
        "finance_5mins",
        "finance_daily",
        "finance_fundamental_data",
        "finance_standard",
        "gov",
        "home",
                       ]:
        try:
            success = mysql_create_database(host="localhost",
                                            port=3306,
                                            username=os.getenv('local_db_username'),
                                            password=os.getenv('local_db_password'),
                                            database_name=finance_db,
                                            )
            if success:
                print(f"Database '{finance_db}' created successfully.")
            else:
                print(f"Failed to create database '{finance_db}'.")
        except Exception:
            pass
        try:
            dagster_logger.info(f"{finance_db} restore starting now")
            success = mysql_import_dump(host="localhost",
                                            port=3306,
                                            username=os.getenv('local_db_username'),
                                            password=os.getenv('local_db_password'),
                                            database_name=finance_db,
                                            dump_file_path=f"{data_path.replace('/','//') + finance_db}_dump_latest.sql"
                                            )
            dagster_logger.info(f"{finance_db} complete in {datetime.datetime.now() - today} ")
            if success:
                print(f"Database '{finance_db}' successfully imported.")
            else:
                print(f"Failed to import database '{finance_db}'.")
        except Exception:
            pass

# POSTGRES
def pg_restore_database(sql_file_path: str, database_name:str,) -> None:
    """
    Uses `psql` to load a .sql file into a PostgreSQL database.
    Creates the database if it doesn't exist.

    1. Check if the database exists
    2. Load the SQL file

    Parameters:
    - sql_file_path (str): Path to the .sql file.
    - database_name (str): Database to create/use.

    Returns:
        None
    """
    if not os.path.exists(sql_file_path):
        raise FileNotFoundError(f"SQL file not found: {sql_file_path}")

    user = os.getenv('local_db_username')
    password = os.getenv('local_db_password')
    host = os.getenv('local_db_address')
    port = os.getenv('local_db_port')
    print(host)

    # Step 1: Load the SQL file
    print(f"Running SQL file '{sql_file_path}' into database '{database_name}'...")
    run_sql_cmd = [rf"PGPASSWORD={password}",
        "psql",
        "-U", user,
        "-h", host,
        "-p", str(port),
        "-d", database_name,
        "-f", sql_file_path
    ]
    execute_cmd_ubuntu_sudo(run_sql_cmd)
    print("SQL file executed successfully.")

def pg_backup_database(database_name: str, host:str, password:str, user:str, port:str, file_location:str = "/home/jay/PycharmProjects/jragbeer_home/data/") -> None:
    """
    Creates a backup (dump) of a PostgreSQL database and saves it to a .sql file.

    This function uses `pg_dump` to export the specified PostgreSQL database to a SQL dump file,
    saving the output to a path constructed as `{data_path}{database_name}_dump_latest.sql`.

    Environment Variables Required:
    - local_db_address: the hostname or IP address of the PostgreSQL server
    - local_db_username: the username for authentication
    - local_db_password: the password for authentication (used via PGPASSWORD)

    Args:
        database_name (str): The name of the PostgreSQL database to back up.
        host (str): The hostname or IP address of the PostgreSQL server
        password (str): The password for the PostgreSQL user
        user (str): The username for the PostgreSQL user
        port (str): The port for the PostgreSQL server

    Logs:
        - Start and completion time of the dump process.
    """
    dagster_logger.info(f"{database_name} dump starting now")

    t = [rf"PGPASSWORD={password}",
        "pg_dump",
        rf"-h{host}",
        rf"-U{user}",
        rf"-d{database_name}",
        rf"-p{port}",
        rf"-f{file_location}pg_{database_name}_dump_latest.sql"
    ]
    dagster_logger.info(t)
    output = execute_cmd_ubuntu_sudo(t)
    if not output:
        dagster_logger.info(f"{database_name} dump failed with SUDO.")
        dagster_logger.info(error_handling())
        execute_cmd_ubuntu_normal(t)
    dagster_logger.info(f"{database_name} completed in {datetime.datetime.now() - today} ")


def sql_table_drop_duplicates_single(new_engine: sqlalchemy.Engine, table_name:str) -> None:
    """
    This function will drop duplicate tables using pandas.
    Args:
        new_engine: The sql engine of the database
        table_name: The name of the table to drop duplicates.

    Returns:
        None
    """
    try:
        tmp_df = pd.read_sql(f'select * from ``{table_name}``', new_engine)
    except Exception:
        try:
            tmp_df = pd.read_sql(f'select * from {table_name}', new_engine)
        except Exception:
            tmp_df = pd.read_sql(f'select * from ```{table_name}```', new_engine)

    print(f"\tTable has {len(tmp_df.index)} rows at start")
    try:
        tmp_df = tmp_df.sort_values('Datetime')
    except Exception:
        try:
            tmp_df = tmp_df.sort_values('Date')
        except Exception:
            pass
    try:
        tmp_df['Datetime'] = tmp_df['Datetime'].apply(lambda x: x.replace(tzinfo=None))
    except Exception:
        try:
            tmp_df['Date'] = tmp_df['Date'].apply(lambda x: x.replace(tzinfo=None))
        except Exception:
            pass
    tmp_df=tmp_df.drop_duplicates()
    print(f"\tTable has {len(tmp_df.index)} rows at end")

    tmp_df.to_sql(f"{table_name}", new_engine, index=False, if_exists='replace')
    print(f'{table_name} replaced')
def sql_table_drop_duplicates(list_of_table_names: list[str], connection_string: str | None =None, database:str | None =None) -> None:
    """
    This function will call the sql_table_drop_duplicates_single on each of the tables provided.
    Either the connection string or the database name must be provided.

    Args:
        list_of_table_names: The names of the tables to drop duplicates.
        connection_string (str): The connection string of the database. Defaults to None.
        database: The name of the database. Defaults to None.

    Returns:
        None
    """
    if connection_string:
        db_url = connection_string
    else:
        if not database:
            raise ValueError("either connection_string or database must be provided.")
        db_url = f"postgresql://{os.getenv('local_db_username')}:{os.getenv('local_db_password')}@{os.getenv('local_db_address')}:{os.getenv('local_db_port')}/{database}"
    new_engine = sqlalchemy.create_engine(db_url)
    print(new_engine)
    for ii, table_name in enumerate(tqdm(list_of_table_names)):
        print(ii, table_name)
        if ii % 25 == 0:
            dagster_logger.info(f'Dropping Duplicates on {table_name}')
        try:
            sql_table_drop_duplicates_single(new_engine, table_name)
        except Exception:
            dagster_logger.info(error_handling())
def db_drop_duplicates(database: str = '5mins', num_splits:int = 4 ) -> None:
    """
    This function will drop duplicates in tables in the database requested. It uses dask and parallel processing to
    split the tables in num_splits groups. Each group is processed by a core via dask.
    Args:
        database: The name of the database to drop duplicates for
        num_splits: The number of splits to use for splitting the data into for parallel processing

    Returns:
        None

    """
    # print time this function starts
    now = datetime.datetime.now()
    dagster_logger.info(now)
    db_url = f"postgresql://{os.getenv('local_db_username')}:{os.getenv('local_db_password')}@{os.getenv('local_db_address')}:{os.getenv('local_db_port')}/{database}"
    # flatten list of tables
    list_of_tables = get_all_sql_table_names(connection_string=db_url)
    process_list_with_dask(list_of_tables, sql_table_drop_duplicates, num_splits, 'distributed',
                           priority=2,
                           kwargs={"database":database})
def copy_backup_to_cloud(database_name: str, local_backup_folder:str) -> None:
    """
    This is a wrapper for adls_upload_file() function that sends the backup .sql file to the cloud storage bucket.
    Args:
        database_name: The name of the database to send to Azure
        local_backup_folder (str): The location of the folder where the backup will be stored
    Returns:
        None
    """
    dagster_logger.info(f"{database_name} migrating to BLOB")
    adls_upload_file(local_backup_folder + f"pg_{database_name}_dump_latest.sql", f"database_backups/{database_name}_dump_latest.sql")
    dagster_logger.info(f"{database_name} migration to BLOB complete.")
def backup_databases(local_backup_folder:str, db: str ='postgres', host:str | None = None, port: int | str| None = None) -> None:
    """
    This function is used to backup the databases. It runs the backup_database() function and then uploads the backup to
    the Azure Blob Storage bucket.

    Args:
        db (str): The name of the database type.
        local_backup_folder (str): The location of the folder where the backup will be stored
        host (str): The hostname/address of the local db to backup

    Returns: None

    """
    if not host:
        host = os.getenv("local_db_address")
    password = os.getenv("local_db_password")
    user = os.getenv('local_db_username')
    if not port:
        port = os.getenv("local_db_port")
    else:
        port = str(port)

    for db_name in [
        "finance_1mins",
        "finance_5mins",
        "finance_daily",
        "finance_fundamental_data",
        "finance_standard",
        "gov",
        "home",
        "sports_betting",
        "sports_nba",
    ]:

        if db == 'mysql':
            mysql_backup_database(db_name)
        else:
            pg_backup_database(db_name, host=host, password=password, user=user, port = port, file_location=local_backup_folder)
        copy_backup_to_cloud(db_name, local_backup_folder)
def get_all_sql_table_names(user=None, hostname=None, password=None, port=None, database=None,
                            connection_string=None) -> list[str]:
    """
    This function returns a list of all SQL table names in the database. It uses sqlalchemy.

    Args:
        user: The username of the database.
        hostname: The hostname of the database.
        password: The password for the database user.
        port: The port to use.
        database: The name of the database to use.
        connection_string: The connection string.

    Returns:
        list: The list of SQL table names.
    """
    if connection_string:
        db_url = connection_string
    else:
        # Build the database URL
        if not all([user, hostname, password, port, database]):
            raise ValueError(
                f"Not all of user={user}, hostname={hostname}, password={password}, port={port}, database={database} are filled.")
        db_url = f"postgresql://{user}:{password}@{hostname}:{port}/{database}"

    # Create a SQLAlchemy engine
    engine = sqlalchemy.create_engine(db_url)

    # Reflect the tables
    metadata = sqlalchemy.MetaData()
    metadata.reflect(bind=engine)

    # Get the table names
    table_names = list(metadata.tables.keys())

    # Close the engine connection
    engine.dispose()
    return table_names





from jragbeer_common.azure import \
    jragbeer_common_azure as jragbeer_common_azure
from jragbeer_common.common import \
    jragbeer_common_data_eng as jragbeer_common_data_eng
from jragbeer_common.dask import jragbeer_common_dask as jragbeer_common_dask
from jragbeer_common.ubuntu import \
    jragbeer_common_ubuntu as jragbeer_common_ubuntu



import os
import time
from pprint import pprint
from typing import Callable, Optional

import dask
import pandas as pd
import paramiko  # type: ignore
import sqlalchemy
from dask.distributed import get_client

from ..common.jragbeer_common_data_eng import (dagster_logger, error_handling,
                                               path)
from ..ubuntu.jragbeer_common_ubuntu import (create_yaml_from_dict,
                                             execute_cmd_ubuntu_sudo,
                                             execute_script_with_cmd,
                                             get_remote_process_ids_ubuntu,
                                             give_write_permission_to_folder,
                                             kill_remote_ubuntu_process_ids)


def update_dask_environment_vars_local(env_dict):
    folder = "/etc/dask"
    execute_cmd_ubuntu_sudo(f"mkdir {folder}")
    give_write_permission_to_folder(folder)
    create_yaml_from_dict(env_dict, "/etc/dask/dask.yaml")

def create_dask_scheduler(hostname:str, username:str, password:str,) -> None:
    # SSH into the remote machine
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username, password=password)

    command = (
        "docker exec finance_scheduler bash -c "
        "'cd /opt/dagster/app && "
        "source .venv/bin/activate && "
        "nohup dask scheduler --host 0.0.0.0 > /tmp/dask_scheduler.log 2>&1 &'"
    )

    stdin, stdout, stderr = ssh.exec_command(command)
    dagster_logger.info(f"stdin:{stdin} | stdout:{stdout} | stderr:{stderr}")
    dagster_logger.info(f"Dask Scheduler created on remote machine : {hostname}")

def create_dask_worker(hostname:str, username:str, password:str, scheduler_ip:str='localhost', nworkers: int=1, mem_limit : str = '1GB') -> None:
    # SSH into the remote machine
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username, password=password)

    command = (
        "docker exec finance_scheduler bash -c "
        "'cd /opt/dagster/app && "
        "source .venv/bin/activate && "
        f"nohup dask worker tcp://{scheduler_ip}:8786 --nthreads 1 --nworkers {nworkers} --memory-limit {mem_limit} > /tmp/dask_worker.log 2>&1 &'"
    )
    stdin, stdout, stderr = ssh.exec_command(command)
    dagster_logger.info(f"stdin:{stdin} | stdout:{stdout} | stderr:{stderr}")
    dagster_logger.info(f"Dask Worker created on remote machine : {hostname} / {nworkers} workers with {mem_limit} mem. each")


def kill_and_redeploy_dask_home_setup() -> None:
    kill_dask_deployment_home_setup()
    time.sleep(10)
    deploy_dask_home_setup()
def deploy_dask_home_setup() -> None:
    create_dask_scheduler(os.getenv('cluster_server_1_address', 'cluster_server_1_address'), os.getenv('cluster_server_1_username', 'cluster_server_1_username'), os.getenv('cluster_server_1_password', 'cluster_server_1_password'),)

    # create worker 1
    create_dask_worker(os.getenv('cluster_server_1_address', 'cluster_server_1_address'), os.getenv('cluster_server_1_username', 'cluster_server_1_username'), os.getenv('cluster_server_1_password', 'cluster_server_1_password'),
                           os.getenv('cluster_server_1_address', 'cluster_server_1_address'), nworkers=11, mem_limit = '3.5GB')
    # create worker 2
    create_dask_worker(os.getenv('cluster_server_2_address','cluster_server_2_address'), os.getenv('cluster_server_2_username', 'cluster_server_2_username'), os.getenv('cluster_server_2_password', 'cluster_server_2_password'),
                           os.getenv('cluster_server_1_address', 'cluster_server_1_address'), nworkers=7, mem_limit = '6GB')
    time.sleep(9)

def kill_dask_deployment_home_setup():
    # server 0 is used for task scheduler and the rest used for dask
    for server_number in [1,2]:
        dagster_logger.info(os.getenv(f'cluster_server_{server_number}_address'))
        pids_with_dask_before = get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
        dagster_logger.info(pids_with_dask_before)
        kill_remote_ubuntu_process_ids(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), pids_with_dask_before)
        time.sleep(5)
        pids_with_dask_after =  get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
        ctr = 0
        while len(pids_with_dask_after) > 2:
            dagster_logger.info(f"There are still 3 processes / {pids_with_dask_after}")
            time.sleep(5)
            pids_with_dask_after =  get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
            ctr = ctr + 1
            if ctr == 3:
                kill_remote_ubuntu_process_ids(os.getenv(f'cluster_server_{server_number}_address'),
                                               os.getenv(f'cluster_server_{server_number}_username'),
                                               os.getenv(f'cluster_server_{server_number}_password'),
                                               pids_with_dask_after)
            if ctr == 6:
                raise RuntimeError

        dagster_logger.info(pids_with_dask_after)
        # for each machine, ensure that there are fewer processes and that only the two grep/ps processes are running
        assert len(pids_with_dask_before) >= len(pids_with_dask_after), "There are more dask tasks now than before"
        assert len(pids_with_dask_after) < 3, "There are more than grep | ps process running dask"

def upload_files_to_dask_cluster():
    with sqlalchemy.create_engine(os.getenv("home_connection_string")).begin() as conn:
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  conn)['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        print(f"Using Dask Cluster: {str(client)}")
        client.upload_file('/home/jay/PycharmProjects/finance/finance_data_eng.py')
        client.upload_file('/home/jay/PycharmProjects/finance/finance_funcs.py')
        print(client)


def process_list_with_dask(input_list: list, func: Callable, num_splits: int, cluster:str = 'distributed', priority:int = 1, kwargs: Optional[dict] = None,) -> None:
    """
    Split the input list into multiple sublists using Dask and execute the provided function on each split list.

    :param input_list: The input list to be split.
    :param func: The function to be executed on each split list.
    :param num_splits:  The number of splits to create from the input list.
    :param cluster:  Either "local" or "distributed". This function chooses the cluster to send the tasks to.
    :param kwargs:  Sometimes the func needs extra kwargs, this should be a dict if present, else None
    :param priority:  priority, an int with default 1. Higher has a higher priority and will run first on the cluster.
    :return: None
    """

    # Calculate the size of each split
    split_size = len(input_list) // num_splits
    split_size = max(1, split_size)

    # Create the splits using Dask
    splits = [input_list[i:i + split_size] for i in range(0, len(input_list), split_size)]
    dagster_logger.info(f"{num_splits} splits each of around {split_size} size made. {len(input_list)} in total.")
    if cluster == 'distributed':
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  sqlalchemy.create_engine(os.getenv('home_connection_string')))['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        dagster_logger.info(str(running_cluster_location))
        dagster_logger.info(f"Using Distributed Dask Cluster : {str(client)}")

    elif cluster == 'local':

        try:
            dagster_logger.info("Checking for Running Local Dask Cluster")
            client = dask.distributed.get_client(f"tcp://{os.getenv('local_db_address')}:8786")
            dagster_logger.info(f"Client acquired at tcp://{os.getenv('local_db_address')}:8786")
            client.shutdown()
            dagster_logger.info("Client shutdown")
            time.sleep(15)
        except Exception:
            dagster_logger.info(error_handling())
            dagster_logger.info("No Client found")

        dagster_logger.info('Creating Local Dask Cluster')
        abc = execute_script_with_cmd(path + "finance_launch_dask_cluster.py")
        dagster_logger.info(str(abc))

        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'local_dask_cluster'""",  sqlalchemy.create_engine(os.getenv('home_connection_string')))['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        dagster_logger.info(str(running_cluster_location))
        dagster_logger.info(f"Using Local Dask Cluster : {str(client)}")

    # Create Dask delayed objects for each split and apply the provided function
    if kwargs:
        delayed_results = [dask.delayed(func)(split, **kwargs) for split in splits]
    else:
        delayed_results = [dask.delayed(func)(split) for split in splits]

    with sqlalchemy.create_engine(os.getenv('home_connection_string')).begin() as conn:
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  conn)['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)

    # Compute the results using Dask's parallel processing capabilities
    output = dask.compute(*delayed_results, priority=priority, )

    return output


def find_number_of_free_dask_workers():
    with sqlalchemy.create_engine(os.getenv('home_connection_string')).begin() as conn:
        running_cluster_location = pd.read_sql(""" SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster' """, conn)['value'].values[0]
        client = get_client(running_cluster_location)
        pprint(f"Using Dask Cluster: {str(client)}")
        abc = client.processing()
        done = [1 for v in abc.values() if len(v) == 0]
        return sum(done)
